
source('config.R')
source('edited_PlotEquiMap.R')

dataset <- 'ERA5land'
accum_scales <- 1:48
max_accums <- c(24,48)
config$lon_min <- 30
config$lon_max <- 37
config$lat_min <- -9
config$lat_max <- -3
config$years <- 1961:2023
config$retrieve <- TRUE
config$regions = list('Tanzania' = list(lon = 35, lat = -5))
config$grid <- 'r3600x1801'
config$figures_path <- '/esarchive/scratch/ptrascas/'

if (dataset %in% config$observations) {
  fyears <- NA
  year_dim <- 'year'
  month_dim <- 'month'
  target_dims <- c(month_dim, year_dim)
  output_dims <- list()
  for (accum_scale in accum_scales){
    output_dims[[paste0('accum',accum_scale)]] <- target_dims
  }
} else {stop('Incorrect dataset')}

## Variables
tasmin <- startR_calls(variable = 'tasmin', dataset = dataset, fyears = fyears, config = config)
tasmax <- startR_calls(variable = 'tasmax', dataset = dataset, fyears = fyears, config = config)
pr <- startR_calls(variable = 'pr', dataset = dataset, fyears = fyears, config = config)

## Longitudes and latitudes
if (identical(tasmin$lat,tasmax$lat) & identical(tasmin$lat,pr$lat) & 
    identical(tasmin$lon,tasmax$lon) & identical(tasmin$lon,pr$lon)) {
  lat <- tasmin$lat
  lon <- tasmin$lon
} else {stop('Different lons or lats for the variables')}

accum <- multiApply::Apply(data = list(tasmin = tasmin$data,
                                       tasmax = tasmax$data,
                                       pr = pr$data,
                                       lat = lat), 
                           target_dims = list(tasmin = target_dims, 
                                              tasmax = target_dims, 
                                              pr = target_dims, 
                                              lat = NULL), 
                           output_dims = output_dims,
                           fun = P_PET_Accum,
                           pet_method = config$pet_method,
                           accum_scales = accum_scales,
                           units_tas = tasmin$units_tas, 
                           units_pr = pr$units_pr, 
                           year_dim = year_dim, 
                           month_dim = month_dim, 
                           ncores = config$ncores)

for (region in names(config$regions)){
  
  ## Indices for selecting the region
  i_lon <- which.min(abs(lon - config$regions[[region]]$lon))
  i_lat <- which.min(abs(lat - config$regions[[region]]$lat))
  
  spei <- array(data = NA, dim = c(month = 12*length(config$years), accum = length(accum_scales)))
  for (accum_scale in accum_scales){
    ## Selecting the region
    aux <- ClimProjDiags::Subset(x = accum[[paste0('accum',accum_scale)]], along = c('lon','lat'), indices = list(i_lon,i_lat), drop = TRUE)
    ## Computing SPEI index
    aux <- SPEI_from_Accum(accum = drop(aux), year_dim = 'year', method = config$spei_method, distribution = config$spei_distribution, fit = config$spei_fit, ncores = config$ncores)
    ## Merging month and year dimensions
    spei[ , which(accum_scales == accum_scale)] <- CSTools::MergeDims(data = aux, merge_dims = c('month','year'), rename_dim = 'month', na.rm = FALSE)
  }
  rm(aux)
  
  ###############################################
  #####----- MULTI TIME SCALE DIAGRAMS -----#####
  ###############################################
  
  names(dim(spei))[which(names(dim(spei)) == c('month','accum'))] <- c('lon','lat')
  
  # Define the palette and generate the colours
  breaks <- seq(-2.5,2.5,0.1)
  palette <- colorRampPalette(c("#d95f0e", "white", "darkgreen"))
  colors <- palette(length(breaks)+1)
  col_inf <- colors[1]
  col_sup <- colors[length(colors)]
  colors <- colors[2:(length(colors)-1)]
  
  for (max_accum in max_accums){
    
    spei2plot <- ClimProjDiags::Subset(x = spei, along = 'lat', indices = 1:max_accum, drop = FALSE)
    lon2plot <- 1:dim(spei2plot)['lon'] #seq(1,10, length.out = dim(spei2plot)['lon'])
    lat2plot <- 1:dim(spei2plot)['lat'] #seq(1,10, length.out = dim(spei2plot)['lat'])
    
    # latmin <- floor(min(lat2plot) / 10) * 10
    # latmax <- ceiling(max(lat2plot) / 10) * 10
    # intxlon <- length(lon2plot) / length(seq(floor(min(lon2plot) / 10) * 10, ceiling(max(lon2plot) / 10) * 10))
    
    edited_PlotEquiMap(var = spei2plot, lon = lon2plot, lat = lat2plot,
                      brks = breaks, triangle_ends = c(T,T), bar_extra_labels = seq(-2.5,2.5,0.2),
                      ylabels = as.character(1:max_accum), intylat = 4,
                      xlabels = paste0(config$years-1), intxlon = 5, axes_tick_scale = 0, axes_label_scale = 0.8,
                      cols = colors, col_inf = col_inf, col_sup = col_sup, 
toptitle = paste('Multi-time scale diagram for the SPEI index', paste0(region,' (lon = ',config$regions[[region]]$lon,'\u00B0, lat = ',config$regions[[region]]$lat,'\u00B0) - ',dataset), sep = '\n'),                                 sizetit = 0.6, colNA = 'grey', filled.continents = F, filled.oceans = F, coast_width = 0.00001, draw_bar_ticks = F,
                      fileout = paste0(config$figures_path,'MultiTimeScale-diagram_',dataset,'_',region,'_SPEI',max_accum,'.png'))
    
  }
  
}
print('Finished')

