
library(ClimProjDiags)
library(CSTools)
library(multiApply)
library(SPEI)
library(TLMoments)
library(lmom)
library(lmomco)

SPEI_from_Accum <- function(accum, method, distribution, fit, year_dim, ncores = NULL){
  
  ## Using the accumulation values calculated with P_PET_Accum.R
  spei <- multiApply::Apply(data = list(acc_data = accum),
                            target_dims = year_dim,
                            output_dims = year_dim,
                            fun = .spX,
                            ref.start = NULL, 
                            ref.end = NULL, 
                            method = method,
                            distribution = distribution, 
                            fit = fit,
                            na.rm = TRUE,
                            ncores = ncores)$output1
  return(spei)
}

.spX <- function(acc_data, ref.start, ref.end, na.rm=TRUE, method = 'parametric', distribution='log-Logistic', fit='ub-pwm'){
  
  if (method=='parametric'){
    spei_mod <- .speiX(acc_data, na.rm=na.rm, method = method, distribution=distribution, fit = fit, ref.start=NULL, ref.end=NULL, x=FALSE)$fitted
  } else if (method=='non-parametric'){
    spei_mod <- .speiX(acc_data, na.rm=na.rm, method = method, distribution=distribution)
  } else {stop('pcX script error: SPEI can be only computed using the following approach: parametric or non-parametric')}
  
  return(drop(spei_mod))
}

.speiX <- function(data, distribution='log-Logistic', fit='ub-pwm', na.rm=TRUE, method = 'parametric',
                   ref.start=NULL, ref.end=NULL, params=NULL, x=FALSE) {
  
  if (!(method %in% c('parametric', 'non-parametric'))) {
    stop('SPEI can be only computed using the following approach: parametric or non-parametric')
  }
  
  if(method=='non-parametric'){
    
    na.rm <- as.logical(na.rm)
    
    if (anyNA(data) && na.rm==FALSE) {
      stop('Error: Data must not contain NAs')
    }
    
    m <- ncol(data)
    n <- length(data)
    
    x <- data
    
    bp = matrix(0,length(x),1)
    for (i in 1:length(x)) {
      bp[i,1]=sum(x[]<=x[i]);  # Writes the rank of the data
    }
    
    SI=(bp-0.44)/(length(x)+0.12)
    
    SPEI=qnorm(SI)
    
    return(SPEI)
    
  }else{
    
    # print('The SPEI is computed using parametric approach')
    
    na.rm <- as.logical(na.rm)
    x <- as.logical(x)
    #if (!exists("data",inherits=F) | !exists("scale",inherits=F)) {
    #	stop('Both data and scale must be provided')
    #}
    if (anyNA(data) && na.rm==FALSE) {
      stop('Error: Data must not contain NAs')
    }
    if (!(distribution %in% c('log-Logistic', 'Gamma', 'PearsonIII'))) {
      stop('Distrib must be one of "log-Logistic", "Gamma" or "PearsonIII"')
    }
    if (!(fit %in% c('max-lik', 'ub-pwm', 'pp-pwm'))) {
      stop('Method must be one of "ub-pwm" (default), "pp-pwm" or "max-lik"')
    }
    if ( (!is.null(ref.start) && length(ref.start)!=2) | (!is.null(ref.end) && length(ref.end)!=2) ) {
      stop('Start and end of the reference period must be a numeric vector of length two.')
    }
    
    if (!is.ts(data)) {
      data <- ts(as.matrix(data), frequency = 1)
    } else {
      data <- ts(as.matrix(data), frequency=frequency(data), start=start(data))
    }
    m <- 1 #ncol(data)
    fr <- frequency(data)
    
    coef = switch(distribution,
                  "Gamma" = array(NA,c(2,m,fr),list(par=c('alpha','beta'),colnames(data),NULL)),
                  "log-Logistic" = array(NA,c(3,m,fr),list(par=c('xi','alpha','kappa'),colnames(data),NULL)),
                  "PearsonIII" = coef <- array(NA,c(3,m,fr),list(par=c('mu','sigma','gamma'),colnames(data),NULL))
    )
    
    dim_one = ifelse(distribution == "Gamma", 2, 3)
    
    if (!is.null(params)) {
      if (dim(params)[1]!=dim_one | dim(params)[2]!=m | dim(params)[3]!=12) {
        stop(paste('parameters array should have dimensions (', dim_one, ', ', m, ', 12)',sep=' '))
      }
    }
    
    # Loop through series (columns in data)
    if (!is.null(ref.start) && !is.null(ref.end)) {
      data.fit <- window(data,ref.start,ref.end)
    } else {
      data.fit <- data
    }
    std <- data*NA
    for (s in 1:m) {
      # Cumulative series (acu)
      acu <- data.fit[,s]
      acu.pred <- data[,s]
      
      # Loop through the months
      for (c in (1:fr)) {
        # Filter month m, excluding NAs
        f <- which(cycle(acu)==c)
        f <- f[!is.na(acu[f])]
        ff <- which(cycle(acu.pred)==c)
        ff <- ff[!is.na(acu.pred[ff])]
        
        # Monthly series, sorted
        month <- sort.default(acu[f], method="quick")
        
        if (length(month)==0) {
          std[f] <- NA
          next()
        }
        
        if (is.null(params)) {
          month_sd = sd(month,na.rm=TRUE)
          if (is.na(month_sd) || (month_sd == 0)) {
            std[f] <- NA
            next
          }
          
          if(distribution != "log-Logistic"){
            pze <- sum(month==0)/length(month)
            month = month[month > 0]
          }
          
          # Stop early and assign NAs if month's data is length < 4
          if(length(month) < 4){
            std[ff,s] = NA
            coef[,s,c] <- NA
            next
          }
          
          # Calculate probability weighted moments based on fit with lmomco or TLMoments
          pwm = switch(fit,
                       "pp-pwm" = pwm.pp(month,-0.35,0, nmom=3),
                       #pwm.ub(month, nmom=3)
                       TLMoments::PWM(month, order=0:2)
          )
          
          # Check L-moments validity
          lmom <- lmomco::pwm2lmom(pwm)
          if ( !are.lmom.valid(lmom) || anyNA(lmom[[1]]) || any(is.nan(lmom[[1]])) ){
            next
          }
          
          # lmom fortran functions need specific inputs L1, L2, T3
          # this is handled by lmomco internally with lmorph
          fortran_vec = c(lmom$lambdas[1:2], lmom$ratios[3])
          
          # Calculate parameters based on distribution with lmom then lmomco
          f_params = switch(distribution,
                            "log-Logistic" = tryCatch(lmom::pelglo(fortran_vec), error = function(e){ parglo(lmom)$para }),
                            "Gamma" = tryCatch(lmom::pelgam(fortran_vec), error = function(e){ pargam(lmom)$para }),
                            "PearsonIII" = tryCatch(lmom::pelpe3(fortran_vec), error = function(e){ parpe3(lmom)$para })
          )
          
          # Adjust if user chose log-Logistic and max-lik
          if(distribution == 'log-Logistic' && fit=='max-lik'){
            f_params = parglo.maxlik(month, f_params)$para
          }
        } else {
          
          f_params = as.vector(params[,s,c])
          
        }
        
        # Calculate cdf based on distribution with lmom
        cdf_res = switch(distribution,
                         "log-Logistic" = lmom::cdfglo(acu.pred[ff], f_params),
                         "Gamma" = lmom::cdfgam(acu.pred[ff], f_params),
                         "PearsonIII" = lmom::cdfpe3(acu.pred[ff], f_params)
        )
        
        std[ff,s] = qnorm(cdf_res)
        coef[,s,c] <- f_params
        
        # Adjust if user chose Gamma or PearsonIII
        if(distribution != 'log-Logistic'){
          std[ff,s] = qnorm(pze + (1-pze)*pnorm(std[ff,s]))
        }
        
      } # next c (month)
      
      ####  Handling infinity is important!! I have made a small trick here. ####
      ## if  Inf <- Takes the maximum 'std' value
      ## if -Inf <- Takes the minimum 'std' value
      
      ###### possibility of adding a parameter to handle Infs
      # std[is.infinite(std) & std < 0] <- min(std[!is.infinite(std)])
      # std[is.infinite(std) & std > 0] <- max(std[!is.infinite(std)])
      
      #std <- drop(std)
      ###
      
    } # next s (series)
    #colnames(std) <- colnames(data)
    
    z <- list(call=match.call(expand.dots=FALSE),
              fitted=std,coefficients=coef,
              distribution=distribution,fit=fit,na.action=na.rm)
    if (x) z$data <- data
    if (!is.null(ref.start)) z$ref.period <- rbind(ref.start,ref.end)
    
    #class(z) <- 'spei'  # Uncomment this line if you want to write all the only fitted values
    return(z)
  }
}
