
library(multiApply)

plot_deterministic_forecast <- function(obs, fcst, title = NULL, xlabs = NULL, ylab = '', ylims = NULL, xlabs_pos = NULL, ylabs_pos = NULL,
                                        style = 'shading', col_obs = 'black', col_hcst = 'blue', col_fcst = 'red', 
                                        member_dim = 'member', time_dim = 'year', legend_text = c('Observations','Hindcast','Forecast'),
                                        na.rm = FALSE, width = 1000, height = 500, res = NA, fileout = NULL){
  
  n_obs <- as.numeric(dim(obs)[time_dim])
  n_fcst <- as.numeric(dim(fcst)[time_dim])
  
  if (is.null(ylims)){ylims <- c(-max(abs(fcst) + 0.1, abs(obs) + 0.1, na.rm = TRUE), max(abs(fcst) + 0.1, abs(obs) + 0.1, na.rm = TRUE))}
  if (is.null(xlabs)){xlabs <- 1:n_fcst}
  if (is.null(xlabs_pos)){xlabs_pos <- -2 + seq(from = 1, to = n_fcst, by = 5)}
  if (is.null(ylabs_pos)){ylabs_pos <- ylims[1] - 0.045 + par('usr')[3]}
  
  ## Opening the output file
  if (!is.null(fileout)){png(filename = fileout, width = width, height = height, res = res)}
  
  ## Empty figure with the labs
  plot(x = 1:n_fcst, y = rep(0,n_fcst), col = col_obs, type = 'l', lwd = 1, xlab = '', ylab = ylab, main = title, 
       xlim = c(1,n_fcst), ylim = ylims, cex.lab = 1.2, cex.main = 1.2, cex.axis = 1.2, frame.plot = TRUE, xaxt = 'n')
  axis(1, at = 1:n_fcst, labels = FALSE)
  text(cex = 1.2, x = xlabs_pos, y = ylabs_pos, labels = xlabs[seq(from = 1, to = n_fcst, by = 5)], srt = 45, pos = 1, xpd = TRUE)
  
  ## Auxiliary lines
  abline(h = 0, lw = 1)
  abline(v = seq(from = 1, to = n_fcst, by = 5), lty = 2, lwd = 1, col = 'grey')
  abline(h = seq(from = floor(ylims[1]), to = ceiling(ylims[2]), by = 1), lty = 2, col = 'grey')
  
  ## Ensemble spread
  if (identical(style,'boxplot')){
    ## Hindcast
    for (y in 1:n_obs){boxplot(x = ClimProjDiags::Subset(x = fcst, along = time_dim, indices = y, drop = 'selected'), at = y, notch = FALSE, range = 0, add = TRUE, col = adjustcolor(col_hcst, alpha.f = 0.60), boxwex = 1, axes = F)}
    ## Forecast
    if (n_fcst > n_obs){
      for (y in (n_obs+1):n_fcst){boxplot(x = ClimProjDiags::Subset(x = fcst, along = time_dim, indices = y, drop = 'selected'), at = y, notch = FALSE, range = 0, add = TRUE, col = adjustcolor(col_fcst, alpha.f = 0.60), boxwex = 1, axes = F)}
    }
  } else if (identical(style,'shading')){
    fcst_quantiles <- multiApply::Apply(data = fcst, target_dims = member_dim, fun = function(x){quantile(x = as.vector(x), type = 8, na.rm = na.rm)}, ncores = 1)$output1
    ## Hindcast
    polygon(x = c(1:n_obs,rev(1:n_obs)), y = c(fcst_quantiles[4,1:n_obs],rev(fcst_quantiles[5,1:n_obs])), col = adjustcolor(col_hcst, alpha.f = 0.10), border = NA)
    polygon(x = c(1:n_obs,rev(1:n_obs)), y = c(fcst_quantiles[2,1:n_obs],rev(fcst_quantiles[4,1:n_obs])), col = adjustcolor(col_hcst, alpha.f = 0.30), border = NA)
    polygon(x = c(1:n_obs,rev(1:n_obs)), y = c(fcst_quantiles[1,1:n_obs],rev(fcst_quantiles[2,1:n_obs])), col = adjustcolor(col_hcst, alpha.f = 0.10), border = NA)
    ## Forecast
    if (n_fcst > n_obs){
      polygon(x = c((n_obs):n_fcst,rev((n_obs):n_fcst)), y = c(fcst_quantiles[4,(n_obs):n_fcst],rev(fcst_quantiles[5,(n_obs):n_fcst])), col = adjustcolor(col_fcst, alpha.f = 0.10), border = NA)
      polygon(x = c((n_obs):n_fcst,rev((n_obs):n_fcst)), y = c(fcst_quantiles[2,(n_obs):n_fcst],rev(fcst_quantiles[4,(n_obs):n_fcst])), col = adjustcolor(col_fcst, alpha.f = 0.30), border = NA)
      polygon(x = c((n_obs):n_fcst,rev((n_obs):n_fcst)), y = c(fcst_quantiles[1,(n_obs):n_fcst],rev(fcst_quantiles[2,(n_obs):n_fcst])), col = adjustcolor(col_fcst, alpha.f = 0.10), border = NA)
    }
  } else {stop('fcst_stype must be either boxplot or shading')}
  
  ## Observations
  lines(x = 1:n_obs, y = obs, col = col_obs, lty = 'solid', lwd = 5)
  
  ## Ensemble mean
  ## Hindcast
  lines(x = 1:n_obs, y = multiApply::Apply(data = fcst, target_dims = member_dim, fun = mean, na.rm = na.rm, ncores = 1)$output1[1:n_obs], col = col_hcst, type = 'l', lwd = 5)
  ## Forecast
  if (n_fcst > n_obs){
    lines(x = (n_obs):n_fcst, y = multiApply::Apply(data = fcst, target_dims = member_dim, fun = mean, na.rm = na.rm, ncores = 1)$output1[(n_obs):n_fcst], col = col_fcst, type = 'l', lwd = 5)
  }
  
  ## Legend with the skill scores
  if (!is.null(legend)){legend(x = 1, y = ylims[2], text.col = c(col_obs, col_hcst, col_fcst), cex = 1.2, legend = legend_text)}
  
  ## Closing the output file
  if (!is.null(fileout)){dev.off()}
  
}
