
library(fields)
library(multiApply)

plot_probabilistic_forecast <- function(obs_probs, fcst_probs, rocss, title = '', cols = c("Blues", "YlOrBr", "Reds"),
                                        xlabs = NULL, style = 'probabilities', cat_dim = 'bin', time_dim = 'year', auxiliar_lines = TRUE,
                                        width = 2000, height = 1000, res = 200, cut_figure = TRUE, fileout = NULL){
  
  n_obs <- as.numeric(dim(obs_probs)[time_dim])
  n_fcst <- as.numeric(dim(fcst_probs)[time_dim])
  if (is.null(xlabs)){xlabs <- 1:n_fcst}
  
  ## Opening the output file
  if (!is.null(fileout)){png(filename = fileout, width = width, height = height, res = res)}
  
  ## Colours
  .tercileBrewerColorRamp <- function(ncolors, cols){
    library(RColorBrewer)
    ncolors <- ncolors-2
    tmp <- brewer.pal(ncolors, cols[1])
    rval <- data.frame(low=c(tmp[1],tmp[1],tmp), stringsAsFactors=F)
    tmp <- brewer.pal(ncolors, cols[2])
    rval$middle <- c(tmp[1],tmp[1],tmp)
    tmp <- brewer.pal(ncolors, cols[3])
    rval$high <- c(tmp[1],tmp[1],tmp)
    return(rval)
  }
  t.color <- .tercileBrewerColorRamp(ncolors = 10, cols = cols)     
  brks <- c(seq(0,1,length=nrow(t.color)+1))
  
  ## Figure
  # opar <- par(no.readonly=TRUE)
  # par(mar = c(4, 5, 0, 3)) #change lines of the margins
  # par(oma = c(0, 0, 0, 3)) #change size of the outer margins (ROCSS on the right)
  par(oma = c(0, 4, 0, 0)) #change size of the outer margins (ROCSS on the left)
  
  ## Ensemble spread (hindcast and forecast)
  if (identical(style,'probabilities')){
    
    ## Color boxes
    image(1:n_fcst, c(-1.5,-0.5), matrix(ClimProjDiags::Subset(x = fcst_probs, along = cat_dim, indices = 1, drop = 'selected')), breaks=brks, col=t.color$low, ylab="", xlab="", asp = 1.5, yaxt="n", bty = "n", axes = FALSE)      
    image(1:n_fcst, c(-0.5,0.5), matrix(ClimProjDiags::Subset(x = fcst_probs, along = cat_dim, indices = 2, drop = 'selected')), breaks=brks, col=t.color$middle, ylab="", xlab="", asp = 1.5, yaxt="n", bty = "n", axes = FALSE, add=TRUE)      
    image(1:n_fcst, c(0.5,1.5), matrix(ClimProjDiags::Subset(x = fcst_probs, along = cat_dim, indices = 3, drop = 'selected')), breaks=brks, col=t.color$high, ylab="", xlab="", asp = 1.5, yaxt="n", bty = "n", axes = FALSE, add=TRUE)
    
    ## Legend
    fields::image.plot(add = TRUE, horizontal = T, smallplot = c(0.20,0.85,0.30,0.32), legend.only = TRUE, breaks = brks, lab.breaks=c(rep("", length(brks))), col = t.color[,3], zlim=c(0,1))
    fields::image.plot(add = TRUE, horizontal = T, smallplot = c(0.20,0.85,0.26,0.28), legend.only = TRUE, breaks = brks, lab.breaks=c(rep("", length(brks))), col = t.color[,2], zlim=c(0,1))
    fields::image.plot(add = TRUE, horizontal = T, smallplot = c(0.20,0.85,0.22,0.24), legend.only = TRUE, breaks = brks, col = t.color[,1], zlim=c(0,1), legend.lab="Probability of the tercile", legend.cex = 0.6) #, legend.args = list(cex = 0.1, cex.lab = 0.1, cex.main = 0.1, cex.sub = 0.1, cex.axis = 0.1))          
    
  } else if (identical(style,'most-likely')){
    
    ## Most likely tercile
    most_likely <- multiApply::Apply(data = fcst_probs, target_dims = cat_dim, output_dims = cat_dim, fun = function(x){y <- rep(-1,3); y[which(x == max(x))] <- 0.8; return(y)}, ncores = 1)$output1
    ## Color boxes
    image(1:n_fcst, c(-1.5,-0.5), matrix(ClimProjDiags::Subset(x = most_likely, along = cat_dim, indices = 1, drop = 'selected')), breaks=brks, col=t.color$low, ylab="", xlab="", asp = 1.5, yaxt="n", bty = "n", axes = FALSE)      
    image(1:n_fcst, c(-0.5,0.5), matrix(ClimProjDiags::Subset(x = most_likely, along = cat_dim, indices = 2, drop = 'selected')), breaks=brks, col=t.color$middle, ylab="", xlab="", asp = 1.5, yaxt="n", bty = "n", axes = FALSE, add=TRUE)      
    image(1:n_fcst, c(0.5,1.5), matrix(ClimProjDiags::Subset(x = most_likely, along = cat_dim, indices = 3, drop = 'selected')), breaks=brks, col=t.color$high, ylab="", xlab="", asp = 1.5, yaxt="n", bty = "n", axes = FALSE, add=TRUE)
    
  } else {stop('fcst_stype must be either probabilities or most-likely')}
  axis(1, at = 1:n_fcst, labels = xlabs, pos = -1.5, cex.axis = 0.6, las = 2) 
  points(x = 1:n_obs, y = -2 + multiApply::Apply(data = obs_probs, target_dims = cat_dim, fun = which.max)$output1, pch = 21, cex = 0.5, bg = "black")
  axis(2, at=-1:1, labels=c(paste0("Below (ROCSS = ",round(rocss$cat1, 2),')'), paste0("Normal (ROCSS = ",round(rocss$cat2, 2),')'), paste0("Above (ROCSS = ",round(rocss$cat3, 2),')')), cex.axis=0.6, las="2")
  
  # ## Adding ROCSS (on the right)
  # axis(4, at=c(-1:1,2.3), labels=c(rocss$cat1, rocss$cat2, rocss$cat3, "ROCSS"), las="2", tick = FALSE, cex.axis=0.8)
  
  ## Auxiliar lines
  if (isTRUE(auxiliar_lines)){
    abline(h = seq(from = -1.5, to = 1.5, by = 1), lw = 1) ## Horizontal lines
    for (i in seq(from = 0.5, to = n_fcst+1, by = 1)){segments(x0 = i, y0 = -1.5, x1 = i, y1 = 1.5, lwd = 1)} ## Vertical lines
  } else if (!isFALSE(auxiliar_lines)){stop('auxiliar_lines must be TRUE or FALSE')}
  
  ## Title
  mtext(title, side=3, line=-5, adj=0, cex=0.6, font=2)
  
  ## Closing the output file
  if (!is.null(fileout)){
    dev.off()
    # Cropping the figure to remove unnecessary margins
    if (isTRUE(cut_figure)){
      if (identical(style,'probabilities')){
        system(paste0('convert -crop 100%x65%+0%+280% +repage ',fileout,' ',fileout,'_aux'))
      } else if (identical(style,'most-likely')){
        system(paste0('convert -crop 100%x40%+0%+280% +repage ',fileout,' ',fileout,'_aux'))
      }
      system(paste0('mv ',fileout,'_aux ',fileout))
    }
  }
  
}

########## Silly data for testing ##########
# fcst_probs <- array(data = runif(500, min=0, max=1), dim = c(year = length(1960:2020), bin = 3))
# obs_probs <- array(data = runif(500, min=0, max=1), dim = c(year = length(1961:2015), bin = 3))
# rocss <- list(cat1 = 0.5, cat2 = 0.7, cat3 = 0.9)
# title <- paste('Title1','Title2', sep = '\n')
# xlabels <- rep('2000-3000', length(1960:2020))
# plot_probabilistic_forecast(obs = obs_probs, fcst = fcst_probs, title = title, xlabs = xlabels, rocss = rocss, style = 'probabilities', fileout = '~/Desktop/probabilities.png')
# plot_probabilistic_forecast(obs = obs_probs, fcst = fcst_probs, title = title, xlabs = xlabels, rocss = rocss, style = 'most-likely', fileout = '~/Desktop/most-likely.png')
